import os
import PyQt5.QtWidgets as Qt# QApplication, QWidget, QMainWindow, QPushButton, QHBoxLayout
import PyQt5.QtGui as QtGui
import PyQt5.QtCore as QtCore
import logging
import ergastirio.utils

graphics_dir = os.path.join(os.path.dirname(__file__), 'graphics')

'''
The panels defined in this module are meant to be ONLY gui for other classes, i.e. their event are directly connected to methods of other objects
do not have their own logger, and they do not produce any logging event. Rather, they are "directly" connected to methods and functionalities
of the experiment object
'''

class acquisition_control():

    def __init__(self, app, mainwindow, parent, experiment):
        # app           = The pyqt5 QApplication() object
        # mainwindow    = Main Window of the application
        # parent        = a QWidget (or QMainWindow) object that will be the parent of this gui
        # experiment    = an experiment() object, whose acquisition will be controlled by this panel

        self.mainwindow = mainwindow
        self.app = app
        self.parent = parent
        self.experiment  = experiment 

    def create_gui(self): 

        hbox = Qt.QHBoxLayout()
        vbox_acquisition = Qt.QVBoxLayout()

        box_acquisition = Qt.QGroupBox() 
        box_acquisition.setTitle(f"Acquisition modality")
        vbox = Qt.QVBoxLayout()
        vbox_acquisition_row1 = Qt.QHBoxLayout()
        vbox_acquisition_row2 = Qt.QHBoxLayout()
        vbox_acquisition_row3 = Qt.QHBoxLayout()

        vbox_trigger = Qt.QVBoxLayout()
        box_trigger = Qt.QGroupBox() 
        box_trigger.setTitle(f"Trigger")
        box_trigger_vbox = Qt.QVBoxLayout()

        self.button_StartPauseContinuousAcquisition = Qt.QPushButton()
        self.button_StartPauseContinuousAcquisition.setToolTip('Start or pause triggered data acquisition. Each time a trigger (either internal or by instruments) is received, \nit takes a single set of acquisitions, with the number of acquisitions and time interval specified here below.') 
        self.button_StartPauseContinuousAcquisition.clicked.connect(self.click_button_StartPauseContinuousAcquisition)

        self.button_SingleAcquisition = Qt.QPushButton("Take single set acquisitions")
        self.button_SingleAcquisition.setToolTip('Triggers a single set of acquisitions, with the number of acquisitions and time interval specified here below.') 
        self.button_SingleAcquisition.clicked.connect(self.click_button_SingleAcquisition)
            
        self.label_NumbAcquisitions = Qt.QLabel("Number of acquisitions per set = ")
        self.spinbox_NumbAcquisitions = Qt.QSpinBox()
        self.spinbox_NumbAcquisitions.setRange(1, 100)
        self.spinbox_NumbAcquisitions.valueChanged.connect(self.value_changed_spinbox_NumbAcquisitions)
        self.box_MakeAverage = Qt.QCheckBox("and average")
        self.box_MakeAverage.stateChanged.connect(self.click_box_MakeAverage)
        self.box_MakeAverage.setToolTip('When checked, the multiple acquisitions are averaged.')

        self.label_AcquisitionRefreshTime= Qt.QLabel("Separated by (s) = ")
        self.edit_AcquisitionRefreshTime = Qt.QLineEdit()
        self.edit_AcquisitionRefreshTime.setText(str(self.experiment.refresh_time_multiple_acquisitions))
        self.edit_AcquisitionRefreshTime.returnPressed.connect(self.press_enter_edit_AcquisitionRefreshTime)
        self.edit_AcquisitionRefreshTime.setAlignment(QtCore.Qt.AlignRight)
        self.edit_AcquisitionRefreshTime.setStyleSheet("background-color: #FFFFFF;")

        vbox_acquisition_row1.addWidget(self.button_StartPauseContinuousAcquisition)
        vbox_acquisition_row1.addWidget(self.button_SingleAcquisition)
        vbox_acquisition_row2.addWidget(self.label_NumbAcquisitions)
        vbox_acquisition_row2.addWidget(self.spinbox_NumbAcquisitions)
        vbox_acquisition_row2.addWidget(self.box_MakeAverage)
        vbox_acquisition_row3.addWidget(self.label_AcquisitionRefreshTime)
        vbox_acquisition_row3.addWidget(self.edit_AcquisitionRefreshTime)
        vbox.addLayout(vbox_acquisition_row1)
        vbox.addLayout(vbox_acquisition_row2)
        vbox.addLayout(vbox_acquisition_row3)
        box_acquisition.setLayout(vbox)
        vbox_acquisition.addWidget(box_acquisition)
        
        self.panelGlobal = Qt.QWidget()
        self.panelGlobal.setStyleSheet("background-color: #D3D3D3;")
        layout = Qt.QHBoxLayout(self.panelGlobal)
        self.radio_TriggerGlobal = Qt.QRadioButton()
        self.radio_TriggerGlobal.setText("Internal")
        self.radio_TriggerGlobal.setStyleSheet("QRadioButton { font: bold;}");
        self.radio_TriggerGlobal.setChecked(True)
        self.radio_TriggerGlobal.value= "global"
        self.radio_TriggerGlobal.clicked.connect(self.click_radio_global_instrument)
        self.radio_TriggerGlobal.setToolTip('In this modality, data from all instruments is acquired at periodic inteverals of time, set by the \'refresh time\'. \nNote: user must check that each instrument is running and refreshing its own data.') 
        self.label_RefreshTime = Qt.QLabel("refresh time (s) = ")
        self.edit_RefreshTime = Qt.QLineEdit()
        self.edit_RefreshTime.setText(str(self.experiment.refresh_time_internal_trigger))
        self.edit_RefreshTime.returnPressed.connect(self.press_enter_edit_RefreshTime)
        self.edit_RefreshTime.setAlignment(QtCore.Qt.AlignRight)
        self.edit_RefreshTime.setStyleSheet("background-color: #FFFFFF;")
        layout.addWidget(self.radio_TriggerGlobal)
        layout.addWidget(self.label_RefreshTime)
        layout.addWidget(self.edit_RefreshTime)
        layout.addStretch(1)

        self.panelTrigger = Qt.QWidget()
        self.panelTrigger.setStyleSheet("background-color: #D3D3D3;")
        layout = Qt.QHBoxLayout(self.panelTrigger)
        self.radio_TriggerInstrument = Qt.QRadioButton()
        self.radio_TriggerInstrument.setText("By Instrument")
        self.radio_TriggerInstrument.setStyleSheet("QRadioButton { font: bold;}");
        self.radio_TriggerInstrument.value= "by_instrument"
        self.radio_TriggerInstrument.clicked.connect(self.click_radio_global_instrument)
        self.radio_TriggerInstrument.setToolTip('In this modality, one instrument is designated as trigger. Whenever that instrument updates its own data, it triggers an acquisition.') 
        self.label_TriggerInstrument = Qt.QLabel(" = ")
        self.combo_TriggerInstruments = Qt.QComboBox()
        self.combo_TriggerInstruments.resize(self.combo_TriggerInstruments.sizeHint())
        self.combo_TriggerInstruments.currentIndexChanged.connect(self.change_combo_TriggerInstruments)
        self.combo_TriggerInstruments.setStyleSheet("background-color: #FFFFFF;")
        self.label_TriggerInstrumentDelay = Qt.QLabel(", delayed  by (s) = ")
        self.edit_TriggerInstrumentDelay = Qt.QLineEdit()
        self.edit_TriggerInstrumentDelay.setText(str(self.experiment.trigger_delay))
        self.edit_TriggerInstrumentDelay.setStyleSheet("background-color: #FFFFFF;")
        self.edit_TriggerInstrumentDelay.returnPressed.connect(self.press_enter_edit_TriggerInstrumentDelay)
        self.edit_TriggerInstrumentDelay.setAlignment(QtCore.Qt.AlignRight)
        layout.addWidget(self.radio_TriggerInstrument)
        layout.addWidget(self.label_TriggerInstrument)
        layout.addWidget(self.combo_TriggerInstruments)
        layout.addWidget(self.label_TriggerInstrumentDelay)
        layout.addWidget(self.edit_TriggerInstrumentDelay)
        layout.addStretch(1)
        
        self.buttongroup_trigger = Qt.QButtonGroup(self.parent)
        self.buttongroup_trigger.addButton(self.radio_TriggerGlobal)
        self.buttongroup_trigger.addButton(self.radio_TriggerInstrument)

        box_trigger_vbox.addWidget(self.panelGlobal)
        box_trigger_vbox.addWidget(self.panelTrigger)
        box_trigger.setLayout(box_trigger_vbox)
        vbox_trigger.addWidget(box_trigger)
        
        hbox.addLayout(vbox_acquisition)  
        hbox.addStretch(1)
        hbox.addLayout(vbox_trigger) 
        
        self.parent.setLayout(hbox) #This line makes sure that all widgest defined so far are assigned to the widget defines in self.parent
        self.parent.resize(self.parent.minimumSize())
        #The widgets in this list will be enabled when we are NOT continuosly acquiring
        self.widgets_enabled_when_not_continous_acquisition = [ self.button_StartPauseContinuousAcquisition,
                                                                self.button_SingleAcquisition,
                                                                self.spinbox_NumbAcquisitions,
                                                                self.box_MakeAverage,
                                                                self.label_NumbAcquisitions,
                                                                self.label_AcquisitionRefreshTime,
                                                                self.edit_AcquisitionRefreshTime,
                                                                self.radio_TriggerGlobal,
                                                                self.label_RefreshTime,
                                                                self.edit_RefreshTime,
                                                                self.label_TriggerInstrument,
                                                                self.label_TriggerInstrumentDelay,
                                                                self.radio_TriggerInstrument,
                                                                self.combo_TriggerInstruments,
                                                                self.edit_TriggerInstrumentDelay]

        #The widgets in this list will be enabled when we are continuosly acquiring
        self.widgets_enabled_when_continous_acquisition = [ self.button_StartPauseContinuousAcquisition ]
        #The widgets in this list will be disabled when we are continuosly acquiring
        self.widgets_disabled_when_continous_acquisition = [self.button_SingleAcquisition,
                                                            self.spinbox_NumbAcquisitions,
                                                            self.box_MakeAverage,
                                                            self.label_NumbAcquisitions,
                                                            self.label_AcquisitionRefreshTime,
                                                            self.edit_AcquisitionRefreshTime,
                                                            self.radio_TriggerGlobal,
                                                            self.label_RefreshTime,
                                                            self.edit_RefreshTime,
                                                            self.label_TriggerInstrument,
                                                            self.label_TriggerInstrumentDelay,
                                                            self.radio_TriggerInstrument,
                                                            self.combo_TriggerInstruments,
                                                            self.edit_TriggerInstrumentDelay
                                                            ]
        self.populate_combo_TriggerInstruments()

        #Initialize part of the GUI by mimicking events
        self.click_radio_global_instrument()
        self.set_pause_continous_acquisition_state()
        self.box_MakeAverage.setChecked(self.experiment.average_acquisition)
        #self.click_box_MakeAverage(state= QtCore.Qt.Checked if self.experiment.average_acquisition==True else False)
        return self

    ### GUI Events Functions
    def click_button_StartPauseContinuousAcquisition(self): 
        if(self.experiment.continous_acquisition == False):
            self.press_enter_edit_RefreshTime()
            self.press_enter_edit_AcquisitionRefreshTime()
            self.press_enter_edit_TriggerInstrumentDelay()
            self.start_continous_acquisition()
        elif (self.experiment.continous_acquisition == True):
            self.pause_continous_acquisition()
        return

    def press_enter_edit_RefreshTime(self):
        refresh_time_internal_trigger = self.edit_RefreshTime.text()
        self.experiment.refresh_time_internal_trigger = refresh_time_internal_trigger #When doing this assignment, the self.experiment.refresh_time_internal_trigger setter will take care of checking if refresh_time is valid, and eventually update the value
        self.edit_RefreshTime.setText(f"{self.experiment.refresh_time_internal_trigger:.3f}") #In case refresh_time_internal_trigger is not valid, this instruction will restore the displayed value to its previous (valid) value
        return True
 
    def press_enter_edit_AcquisitionRefreshTime(self):
        acquisition_refresh_time = self.edit_AcquisitionRefreshTime.text()
        self.experiment.refresh_time_multiple_acquisitions = acquisition_refresh_time #When doing this assignment, the self.experiment.refresh_time setter will take care of checking if refresh_time is valid, and eventually update the value
        self.edit_AcquisitionRefreshTime.setText(f"{self.experiment.refresh_time_multiple_acquisitions:.3f}") #In case refresh_time is not valid, this instruction will restore the displayed value to its previous (valid) value
        return True

    def click_button_SingleAcquisition(self):
        self.press_enter_edit_AcquisitionRefreshTime()
        self.experiment.trigger_single_set_acquisition()

    def value_changed_spinbox_NumbAcquisitions(self):
        self.experiment.numb_acquisitions_per_trigger = self.spinbox_NumbAcquisitions.value()

    def click_radio_global_instrument(self):
        if self.radio_TriggerGlobal.isChecked():
            self.experiment.set_trigger_modality('internal')
            self.press_enter_edit_RefreshTime()
            ergastirio.utils.enable_widget([self.edit_RefreshTime,self.label_RefreshTime])
            ergastirio.utils.disable_widget([self.combo_TriggerInstruments,self.edit_TriggerInstrumentDelay,self.label_TriggerInstrument,self.label_TriggerInstrumentDelay])
        if self.radio_TriggerInstrument.isChecked():
            self.experiment.set_trigger_modality('external')
            self.change_combo_TriggerInstruments()
            ergastirio.utils.disable_widget([self.edit_RefreshTime,self.label_RefreshTime])
            ergastirio.utils.enable_widget([self.combo_TriggerInstruments,self.edit_TriggerInstrumentDelay,self.label_TriggerInstrument,self.label_TriggerInstrumentDelay])

    def change_combo_TriggerInstruments(self):
        self.press_enter_edit_TriggerInstrumentDelay()

    def press_enter_edit_TriggerInstrumentDelay(self):
        trigger_delay = self.edit_TriggerInstrumentDelay.text()
        if self.radio_TriggerInstrument.isChecked():
            check = self.experiment.set_trigger_instrument(self.combo_TriggerInstruments.currentText(),delay=trigger_delay)
            if check:
                self.edit_TriggerInstrumentDelay.setText(str(self.experiment.trigger_delay))

    def click_box_MakeAverage(self, state):
        if state == QtCore.Qt.Checked:
            status_bool = True
        else:
            status_bool = False
        print(status_bool)
        self.experiment.set_average_modality(status_bool)

    ### END GUI Events Functions

    def populate_combo_TriggerInstruments(self):
        list_instruments = [instrument['fullname'] for instrument in self.experiment.instruments]
        self.combo_TriggerInstruments.clear()
        self.combo_TriggerInstruments.addItems(list_instruments)  

    def start_continous_acquisition(self):

        self.press_enter_edit_RefreshTime() #We emulate this event to make sure to update the refresh time value
        self.experiment.start_continous_acquisition()
        self.set_continous_acquisition_state() # Change some widgets
        return

    def pause_continous_acquisition(self):
        self.experiment.stop_continous_acquisition()
        self.set_pause_continous_acquisition_state() # Change some widgets
        return

    def set_pause_continous_acquisition_state(self):
        #Changes the GUI based on the state
        self.button_StartPauseContinuousAcquisition.setText('Start triggered acquisition')
        ergastirio.utils.enable_widget(self.widgets_enabled_when_not_continous_acquisition)
        self.click_radio_global_instrument() #mimick a click on radio button to reset the enabled/disabled state of the corresponding widgets

    def set_continous_acquisition_state(self):
        #Changes the GUI based on the state
        self.button_StartPauseContinuousAcquisition.setText('Pause')
        ergastirio.utils.enable_widget(self.widgets_enabled_when_continous_acquisition)
        ergastirio.utils.disable_widget(self.widgets_disabled_when_continous_acquisition)

class data_management():
    def __init__(self, app, mainwindow, parent, experiment):
        # app           = The pyqt5 QApplication() object
        # mainwindow    = Main Window of the application
        # parent        = a QWidget (or QMainWindow) object that will be the parent of this gui
        # experiment    = an experiment() object, whose acquisition will be controlled by this panel
        self.mainwindow = mainwindow
        self.app = app
        self.parent = parent
        self.experiment  = experiment 

    def create_gui(self): 
        hbox1 = Qt.QHBoxLayout()

        self.button_SaveData = Qt.QPushButton("Save data")
        self.button_SaveData.setToolTip('Save all currently stored data in a .csv file.') 
        self.button_SaveData.clicked.connect(self.click_button_SaveData)
        self.button_DeleteAllData = Qt.QPushButton("Delete all data")
        self.button_DeleteAllData.setToolTip('Delete all currently stored data.') 
        self.button_DeleteAllData.clicked.connect(self.click_button_DeleteAllData)
        self.button_DeleteLastRowData = Qt.QPushButton("Delete last row")
        self.button_DeleteLastRowData.setToolTip('Delete last row of stored data.') 
        self.button_DeleteLastRowData.clicked.connect(self.click_button_DeleteLastRowData)

        hbox1.addWidget(self.button_SaveData)
        hbox1.addWidget(self.button_DeleteAllData)
        hbox1.addWidget(self.button_DeleteLastRowData)
        hbox1.addStretch(1)

        vbox = Qt.QVBoxLayout()
        vbox.addLayout(hbox1)  
        vbox.addStretch(1)

        self.parent.setLayout(vbox) #This line makes sure that all widgest defined so far are assigned to the widget defines in self.parent
        self.parent.resize(self.parent.minimumSize())

        return self

    ### GUI Events Functions

    def click_button_SaveData(self): 
        filename, _ = Qt.QFileDialog.getSaveFileName(self.mainwindow, 
                        "Save File", "", "Csv Files (*.csv);;Text Files (*.txt)")
        if filename:
            self.experiment.save_stored_data(filename)
        return

    def click_button_DeleteAllData(self):
        answer = Qt.QMessageBox.question(self.parent,'', "Are you sure?", Qt.QMessageBox.Yes | Qt.QMessageBox.No, Qt.QMessageBox.No)
        if answer == Qt.QMessageBox.Yes:
            self.experiment.delete_current_data()
        return

    def click_button_DeleteLastRowData(self):
        self.experiment.delete_row_from_data(-1)
        return


import os
import PyQt5
dirname = os.path.dirname(PyQt5.__file__)
plugin_path = os.path.join(dirname, 'plugins', 'platforms')
os.environ['QT_QPA_PLATFORM_PLUGIN_PATH'] = plugin_path
import PyQt5.QtWidgets as Qt# QApplication, QWidget, QMainWindow, QPushButton, QHBoxLayout
import PyQt5.QtGui as QtGui
import PyQt5.QtCore as QtCore
import logging
import sys
import argparse
import importlib
import numpy as np
import inspect
from importlib.metadata import version  

from ergastirio.experiment import experiment
import ergastirio.utils

class MainWindow(Qt.QMainWindow):
    '''
    The main window contains several menus and a QMdiArea object, 
    The QMdiArea object in turn contains subwindows where instruments, plots, data table and logs will be created
    Object creation:
        window = MainWindow()
    The QMdiArea object is saved as attribute:
        window.mdi 
    The mdi object is also the parent object which needs to be passed as parameter when creating an instance of the experiment object
    For example:
        Experiment = experiment(app,window,window.mdi,config_file)
    The mdi object contains a dictionary, window.mdi.containers whose elements represent the different subwindows.
    Specifically, window.mdi.containers = {'logging':{...},'tabledata':{...},'instruments':{...},'plots':{...} }
    The object
        window.mdi.containers[name]['container']
    is the widget that will act as container for the different part of the gui. 
    Each of this widget is a child (via setWidget() method) of a corresponding QScrollArea object, 
        window.mdi.containers[name]['scrollarea']
    In turns, the scrollarea objects are children (via setWidget() method) of corresponding QMdiSubWindow objects
        window.mdi.containers[name]['subwindow']
    '''
    _views =  {'View 1':{
                "instruments":  [ "2/3", "0", "1/3", "2/3" ],
                "plots":        [ "0", "0", "2/3", "2/3" ],
                "tabledata":    [ "0", "2/3", "2/3", "1/3" ],
                "logging":      [ "2/3", "2/3", "1/3", "1/3" ]
                }
               ,
               'View 2':{
                "instruments":  [ "1/2", "0", "1/2", "1/2"],
                "plots":        [ "0", "0", "1/2", "1/2" ],
                "tabledata":    [ "0", "1/2", "1/2", "1/2" ],
                "logging":      [ "1/2", "1/2", "1/2", "1/2" ]
                }
               ,
               'View 3':{
                "instruments":  [ "2/3", "0", "1/3", "1/2"],
                "plots":        [ "0", "0", "2/3", "1/2" ],
                "tabledata":    [ "0", "1/2", "2/3", "1/2" ],
                "logging":      [ "2/3", "1/2", "1/3", "1/2" ]
                }
               ,
               'View 4':{
                "instruments":  [ "2/3", "0", "1/3", "1/2"],
                "plots":        [ "0", "0", "2/3", "3/5" ],
                "tabledata":    [ "0", "3/5", "2/3", "2/5" ],
                "logging":      [ "2/3", "1/2", "1/3", "1/2" ]
                }
               }
    _default_view = 'View 4'
    _mdi_panels ={  'logging':{'title':'Logging'},
                    'tabledata':{'title':'Data acquisition'},
                    'instruments':{'title':'Instruments'},
                    'plots':{'title':'Plots'}
                      }

    def __init__(self):
        super().__init__()
        self.setWindowTitle(f"{__package__} (v{version(__package__)})")
        self.mdi = Qt.QMdiArea()
        mdi_panels = self._mdi_panels
        self.setCentralWidget(self.mdi)

        #This dictionary will be used to create the subwindows of the mdi, and also to populate the "View" menu
  
        self.current_view = self._views[self._default_view]

        bar = self.menuBar()

        file = bar.addMenu("File")
        file.addAction("New Experiment")

        view = bar.addMenu("View")
        for k in self._views.keys():
            view.addAction(k)
        view.triggered[Qt.QAction].connect(self.action_view)
        
        showhide = bar.addMenu("Show/Hide")
        for mdi_panel in mdi_panels.values():
            showhide.addAction(mdi_panel['title'])
        showhide.triggered[Qt.QAction].connect(self.action_showhide)
        
        #Create the different subdiwndows, based on the element of the dictionary mdi_panels
        for key,mdi_panel in mdi_panels.items():
            mdi_panels[key]['subwindow'] = Qt.QMdiSubWindow()
            mdi_panels[key]['subwindow'].setWindowFlags(QtCore.Qt.CustomizeWindowHint)
            mdi_panels[key]['subwindow'].setWindowTitle(mdi_panels[key]['title'])
            mdi_panels[key]['scrollarea'] = Qt.QScrollArea(self)
            #mdi_panels[key]['scrollarea'].verticalScrollBar().rangeChanged.connect(lambda : self.adjust_position_scroll_area(mdi_panels[key]['scrollarea']))
            #mdi_panels[key]['scrollarea'].verticalScrollBar().rangeChanged.connect(lambda: mdi_panels[key]['scrollarea'].verticalScrollBar().setValue(mdi_panels[key]['scrollarea'].verticalScrollBar().maximum()))
            mdi_panels[key]['container'] = Qt.QWidget(self)
            mdi_panels[key]['subwindow'].setWidget(mdi_panels[key]['scrollarea'])
            mdi_panels[key]['scrollarea'].setWidget(mdi_panels[key]['container'])
            mdi_panels[key]['scrollarea'].setWidgetResizable(True)
            self.mdi.addSubWindow(mdi_panels[key]['subwindow'])
            mdi_panel['container'].show()
        self.mdi.containers = mdi_panels

        self.set_view(self._views[self._default_view])

    def action_view(self, p):
        for k in self._views.keys():
            if p.text() == k:
                self.current_view = self._views[k]
                self.set_view(self._views[k])
        return

    def action_showhide(self,p):
        for mdi_panel in self.mdi.containers.values():
            if p.text() == mdi_panel['title']:
                mdi_panel['subwindow'].setHidden(not mdi_panel['subwindow'].isHidden())

    def resizeEvent(self, event):
        Qt.QMainWindow.resizeEvent(self, event)
        if hasattr(self,'current_view'):
            self.set_view(self.current_view)
        
    def closeEvent(self, event=None):
        if self.experiment:
            self.experiment.close_experiment()

    def set_view(self,view):
        geomMDI = self.mdi.geometry()
        MDI_width = geomMDI.width()
        MDI_height = geomMDI.height()
        MDI_x = geomMDI.x()
        MDI_y = geomMDI.y()
        for panel,geom in view.items():
            geom_abs = [] 
            geom_abs.append(    int(ergastirio.utils.convert_fraction_to_float(geom[0])*MDI_width)  )
            geom_abs.append(    int(ergastirio.utils.convert_fraction_to_float(geom[1])*MDI_height) )
            geom_abs.append(    int(ergastirio.utils.convert_fraction_to_float(geom[2])*MDI_width)  )
            geom_abs.append(    int(ergastirio.utils.convert_fraction_to_float(geom[3])*MDI_height) )
            self.mdi.containers[panel]['subwindow'].setGeometry(*geom_abs)

def main():
    parser = argparse.ArgumentParser(description = "",epilog = "")
    parser.add_argument('-e', 
                        help=f"Path of .json file contaning the configuration of this experiment",
                        action="store", dest="config", type=str, default=None)
    parser.add_argument("-s", "--decrease_verbose", help="Decrease verbosity.", action="store_true")
    args = parser.parse_args()
    
    app = Qt.QApplication(sys.argv)
    window = MainWindow()
    
    app.aboutToQuit.connect(window.closeEvent) 

    if args.config:
        config_file = os.path.abspath(args.config)
    else:
        folder_default_file = os.path.dirname(os.path.abspath(__file__))
        config_file = os.path.join(folder_default_file,'config_default.json')

    Experiment = experiment(app,window,window.mdi,config_file)

    window.show()
    app.exec()# Start the event loop.

if __name__ == '__main__':
    main()


'''
This module contains several function which are used to initialize a certain experiment. 
Before using the functions in this module, an experiment() object needs to be created (typically by the main.pyw file), and a config file needs to be assigned to 
the experiment. After the experiment have been assigned a config file, the function set_up_experiment(exp) is called, where exp is the experiment() object.
'''
import PyQt5.QtWidgets as Qt# QApplication, QWidget, QMainWindow, QPushButton, QHBoxLayout
import PyQt5.QtGui as QtGui
import PyQt5.QtCore as QtCore
import logging
import numpy as np
import importlib
from importlib.metadata import version  
import ergastirio.panels
import ergastirio.widgets
import ergastirio.utils

style1 = """{
            font: bold;
            border: 1px solid black;
            border-radius: 6px;
            margin-top: 6px;
        }
"""

style2 = """::title {
            subcontrol-origin: margin;
            left: 7px;
            padding: -4px 5px 0px 5px;
        }
        """

def setup(exp):
    if not set_up_experiment(exp):
        return False
    if not set_up_experiment_gui(exp):
        return False
    link_data_to_gui(exp)
    return True
  
def set_up_experiment(exp):
    '''
    This function initializes a certain experiment (passed as an input argument) based on the settings stored in the dictionary exp.config.
    It subsequently calls several functions to load instruments, initialize storage arrays, validate user configs
    '''

    if not load_general_settings(exp):
        return False
        
    if not load_instruments(exp):       #Load all instruments based on the strings specified in the list exp.config['Instruments']
        return False

    if not generate_data_headers(exp):  #Based on the loaded instruments, and on the data created by each instrument, it populates the list of strings exp.data_headers
        return False

    if not load_and_check_postprocessed_data(exp):
        return False

    initialize_storage_variable(exp)    #Create Storage variables
    initialize_temporary_variable(exp)
       
    if not validate_plots_settings(exp): #check if the plots settings specified in the .json file are valid (i.e. any quantity defined in 'x' and 'y' is a valid device data
        return False
    return True


def set_up_experiment_gui(exp):
    '''
    This function initializes the GUI of an experiment (passed as an input argument). It must be called after calling the function set_up_experiment()
    '''
    create_gui_table_data(exp,exp.parent.containers['tabledata']['container'])
    create_gui_plots(exp,exp.parent.containers['plots']['scrollarea'])
    if not create_gui_instruments(exp,exp.parent.containers['instruments']['container']): #Create GUIs for all loaded instruments 
        return False
    #
    if 'DefaultView' in exp.config.keys():
        QtCore.QTimer.singleShot(1, lambda : exp.mainwindow.set_view(exp.config['DefaultView']))

        
    return True



def load_general_settings(exp):
    if 'General_Settings' in exp.config.keys():
        exp.logger.info(f"Loading general settings...")
        if 'refresh_time_internal_trigger' in exp.config['General_Settings'].keys():
            exp.refresh_time_internal_trigger = exp.config['General_Settings']['refresh_time_internal_trigger']
        if 'refresh_time_multiple_acquisitions' in exp.config['General_Settings'].keys():
            exp.refresh_time_multiple_acquisitions = exp.config['General_Settings']['refresh_time_multiple_acquisitions']
        if 'average_acquisition' in exp.config['General_Settings'].keys():
            exp.average_acquisition = bool(exp.config['General_Settings']['average_acquisition'])
    #"continous_acquisition": 0,
    #"triggered_acquisition": 1,
    #"refresh_time_multiple_acquisitions": 0.2,
    #"average_acquisition": 1,
    #"trigger_delay": 0,
    #"trigger_instrument": "dev2_pyThorlabsPM100x"
    return True

def load_instruments(exp):
    ''' Loads all instruments defined in exp.config['Instruments'] and stores them in the list exp.instruments
        For each intruments, the respective interface object is also created, but not the gui.

        Returns true if all instruments were loaded correctly, false otherwise
    '''
    exp.logger.info(f"Loading all instruments specified in {exp.config_file}...")
    exp.instruments = []
    i=0
    for instrument in exp.config['Instruments']:
        IsValidPackage = importlib.util.find_spec(instrument) is not None
        if not(IsValidPackage):
            exp.logger.error(f"{instrument} is not a valid python package. Fix the error and restart this application")
            return False
        device_module =  importlib.import_module(instrument+'.main')
        if not(hasattr(device_module,'interface')):
            exp.logger.error(f"{instrument} is a valid package, but it does not contain an 'interface' class.")
            return False
        fullname = f"dev{i}_{instrument}"
        try:
            exp.instruments.append({'name':f"dev{i}",
                                        'fullname':fullname,
                                        'type':instrument,
                                        'interface_class':device_module.interface,
                                        'interface':  device_module.interface(  app=exp.app,
                                                                                mainwindow=exp.mainwindow,
                                                                                name_logger=fullname) #It might make more sense to create the interface objects in the load_instruments function instead
                                    }
                                    )
        except Exception as e:
            exp.logger.error(f"An error occurred while loading the instrument {instrument}: {e}")
            return False
        exp.logger.info(f"Instrument loaded: {exp.instruments[i]['name']} -> {exp.instruments[i]['type']}")
        i = i+1
    return True

def generate_data_headers(exp):
    ''' Generates a list of data headers for the experiment exp, based on the loaded instruments.
    For each instrument in exp.instruments, the list of output data is contained in the keys of the dictionary instrument.['interface_class'].output.
    The headers 'timestamp' and 'time' are also added at the beginning of the data headers.

    Returns True
    '''
    exp.logger.info(f"Preparing the data headers...")
    exp.data_headers =[]
    exp.data_headers.append('timestamp')
    exp.data_headers.append('time')
    for instrument in exp.instruments:
        if hasattr(instrument['interface_class'],'output'):
            for key in instrument['interface_class'].output.keys():
                exp.data_headers.append(instrument['name']+'.'+key)
        else:
            exp.logger.info(f"Instrument {instrument['fullname']} does not have an \'output\' dictionary defined. I wil assume that it does not produce data.")
    exp.logger.info(f"The following data will be acquired: {exp.data_headers}")
    return True

def load_and_check_postprocessed_data(exp):
    ''' Load the post-processed data defined in exp.config['PostProcessedData'] (if any is defined)
        and checks that each of them represents a valid formula.
        It apppen the name of each post-processed data to the list exp.data_headers.

        TO DO: check that no post-processed data has same name as another valid data.

        Returns true if all post-processed data are valid, false otherwhise.
    '''
    if ('PostProcessedData' in exp.config.keys()):
        exp.logger.info(f"Validating post-processed data...")
        for post_processed_data_name,post_processed_data in exp.config['PostProcessedData'].items():
            (msg,flag) = ergastirio.utils.check_list_is_valid_formula(post_processed_data,valid_variable_names=exp.data_headers)
            if not(flag==1):
                exp.logger.error(f"A problem occurred with the post-processed data: {msg}.")
                return False
            exp.data_headers.append(post_processed_data_name)

    return True
    

def initialize_storage_variable(exp):
    ''' Initialize all arrays for data storage. '''      
    exp.data = ergastirio.EnhancedList([]) #np.empty([0,len(exp.data_headers)]) 
    exp.data_std = ergastirio.EnhancedList([])
    return True
    
def initialize_temporary_variable(exp):
    ''' These variables will only be used as temporary arrays for averaging purposes. The number of columns in the storage matrices is given by the length of exp.data_headers. '''
    exp.numb_acq_to_average = 0 # This will contain the number of acquisitions to be averaged, chosen by the user
    exp.data_being_averaged = False #It's a flag which keeps track of whether data is being currently averaged
    exp.data_temp = np.empty([0,len(exp.data_headers)-2]) #The number of columns of expdata_temp is 2 less than the number of columns of exp.data
                                                                #because the first 2 columns of exp.data are for time and acquisition number
    return True

def validate_plots_settings(exp):
    '''
    This function looks at the content of exp.config['Plots'] and check that all settings specified by the user are valid
    '''
    exp.logger.info(f"Validating all plot settings...")
    exp.plots = []
    valid_data = exp.data_headers.copy()
    valid_data.append("acq#") #Need to implement this better
    for plotindex, plot in enumerate(exp.config['Plots']):
        if isinstance(plot,dict) and ('x' in plot.keys()) and ('y' in plot.keys()):
            if not plot['x'] in valid_data:
                exp.logger.error(f"Plot #{plotindex}: {plot['x']} is not a valid device data")
                return False
            if isinstance(plot['y'],str):
                plot['y'] = [plot['y']]
            for y_data in plot['y']:
                if not y_data in valid_data:
                    exp.logger.error(f"Plot #{plotindex}: {y_data} is not a valid device data")
                    return False            
        else:
            exp.logger.error(f"Each plot must be defined in the .json file as a dictionary containing an 'x' and a 'y' key.")
            return False
        exp.plots.append({'name':f"plot{plotindex}",
                            'x':plot['x'],
                            'y':plot['y']})
        exp.logger.info(f"Plot #{plotindex}, x = {plot['x']}, y = {plot['y']}")
    return True

def create_gui_logging(exp,container):
    '''
    Create GUIs for logger
    container
        QWidget which will contain the GUIs
    '''
    exp.logger.info(f"Creating GUI for logger...")
    box = Qt.QVBoxLayout()
    exp.logging_text_area = ergastirio.widgets.LoggerTextArea()
    exp.logging_text_area.add_logger(logger=logging.getLogger(exp.name_logger))
        
    box.addWidget(exp.logging_text_area)
    container.setLayout(box)

    return True

def create_gui_instruments(exp,container):
    '''
    Create GUIs for all loaded instruments, and for the acquisition control panel

    container
        QWidget which will contain the GUIs
    '''

    if ('Alignment_Instruments_Window' in exp.config.keys())  and exp.config['Alignment_Instruments_Window']=='H':
        box = Qt.QHBoxLayout()
    else: 
        box = Qt.QVBoxLayout()

    exp.logger.info(f"Creating GUI for all loaded instruments...")
    
    for i,instrument in enumerate(exp.instruments):
        exp.instruments[i]['frame'] = Qt.QGroupBox() # Qt.QWidget()
        exp.instruments[i]['frame'].setObjectName(exp.instruments[i]['name']+'_frame')
        string = "QGroupBox#"+ exp.instruments[i]['frame'].objectName()
        exp.instruments[i]['frame'].setStyleSheet(string + style1 + string + style2) #This line changes the style of ONLY this QWdiget

        exp.instruments[i]['frame'].setTitle(f"{exp.instruments[i]['name']} ({exp.instruments[i]['type']} v{version(exp.instruments[i]['type'])})")
        
        exp.logging_text_area.add_logger(logger=logging.getLogger(exp.instruments[i]['fullname'])) #Connect the logger of this instrument to the text area for logging in the GUI
        exp.instruments[i]['interface'].create_gui(parent=exp.instruments[i]['frame'])
        exp.instruments[i]['frame'].resize(exp.instruments[i]['frame'].sizeHint())
        box.addWidget(exp.instruments[i]['frame'] )
    box.addStretch(1)

    container.setLayout(box) #This line makes sure that all widgest defined so far are assigned to the widget defined in container
    exp.container_instruments = container
    exp.container_instruments.resize(exp.container_instruments.sizeHint())
    return True

def create_gui_plots(exp,container):
    '''
    Create GUIs for all plots

    container
        QWidget which will contain the GUIs
    '''
    exp.logger.info(f"Creating GUI for all plots...")

    container.mdi = Qt.QMdiArea()
    container.setWidget(container.mdi)
    for plotindex, plot in enumerate(exp.plots):
        exp.plots[plotindex]['subwindow'] = Qt.QMdiSubWindow()
        exp.plots[plotindex]['subwindow'].setWindowFlags(QtCore.Qt.FramelessWindowHint)
        exp.plots[plotindex]['subwindow'].setWindowTitle(exp.plots[plotindex]['name'])
        exp.plots[plotindex]['scrollarea'] = Qt.QScrollArea()
        exp.plots[plotindex]['widget'] = Qt.QWidget()
        exp.plots[plotindex]['subwindow'].setWidget(exp.plots[plotindex]['scrollarea'])
        exp.plots[plotindex]['scrollarea'].setWidget(exp.plots[plotindex]['widget'])
        exp.plots[plotindex]['scrollarea'].setWidgetResizable(True)
        container.mdi.addSubWindow(exp.plots[plotindex]['subwindow'])
        exp.plots[plotindex]['plotobject'] = ergastirio.widgets.PlotObject(exp.app, 
                                                                           exp.mainwindow, 
                                                                           exp.plots[plotindex]['widget'],
                                                                           data_headers = exp.data_headers,
                                                                           plot_config=exp.plots[plotindex])
        exp.plots[plotindex]['widget'].show()
    container.mdi.tileSubWindows()
    exp.container_plots = container
    return True

def create_gui_table_data(exp,container):
    '''
    Create GUIs for the table showing the currently stored data

    container
        QWidget which will contain the GUIs
    '''  
    exp.logger.info(f"Creating GUI for table...")

    layout = Qt.QVBoxLayout()

    ## CREATE ACQUISITION CONTROL PANEL
    exp.logger.info(f"Creating GUI for acquisition control panel...")
    exp.control_panel_container = Qt.QGroupBox() 
    exp.control_panel_container.setObjectName('acquisition_panel')
    string = 'QGroupBox#acquisition_panel'
    exp.control_panel_container.setStyleSheet(string + style1 + '\nQGroupBox' + style2) #This line changes the style of ONLY this QWdiget
    exp.control_panel_container.setTitle(f"Acquisition Panel")
    exp.control_panel = ergastirio.panels.acquisition_control(  exp.app,
                                                                exp.mainwindow,
                                                                exp.control_panel_container,
                                                                exp
                                                                )
    exp.control_panel.create_gui()
    ####

    ## CREATE DATA MANAGEMENT PANEL
    exp.data_management_panel_container = Qt.QGroupBox()
    exp.data_management_panel = ergastirio.panels.data_management(  exp.app,
                                                                    exp.mainwindow,
                                                                    exp.data_management_panel_container,
                                                                    exp)
    exp.data_management_panel.create_gui()
    exp.tabledata = ergastirio.widgets.Table()
    exp.tabledata.data_headers = exp.data_headers
    exp.tabledata.data = exp.data
    ####
    
    layout.addWidget(exp.control_panel_container)
    layout.addWidget(exp.data_management_panel_container)
    layout.addWidget(exp.tabledata,stretch=1)
    container.setLayout(layout)
    exp.container_tabledata = container
    return True

def link_data_to_gui(exp):
    '''
    The data collected in the experiment is stored in exp.data, which is an instance of the EnhancedList class.
    This is a 'dynamic' list, wich allows 'linked objects'. Whenever any data stored in an EnhancedList is changed, any linked
    object is notified, and a copy of the currently stored date is sent to the object.
    This is used as an elegant way to keep the table data and plots always syncronized with the stored data.
    To link an object, we use the method add_syncronized_objects defined in the EnhancedList class

    data.add_syncronized_objects([ InstanceOfTargetObject,  TargetClassProperty])

    Every time that the content of data is changed (e.g. a row is added to acquired data), the object data is also copied into
    the propery InstanceOfTargetObject.TargetClassProperty
    '''
    exp.data.add_syncronized_objects([  exp.tabledata,  ergastirio.widgets.Table.data])
    for plot in exp.plots:
        exp.data.add_syncronized_objects([  plot['plotobject'],  ergastirio.widgets.PlotObject.data])



import ergastirio.experiment_initializer
import PyQt5.QtCore as QtCore
import logging
import json
import numpy as np
import datetime
import ergastirio.utils

class experiment():
    _config_file = ''
    _verbose = True #Keep track of whether this instance of the interface should produce logs or not
    _name_logger = ''
    data_headers =[]    #List of strings, will contain the label of each 'column" of acquired data, based on the instruments connected, in the format "dev#i_dataname#j"
                        # where #i runs from 0 to the number of instruments minus 1, and dataname#j is the name of the jth data created by the i-th instrument.
                        # the data created by each instrument are specified by the keys of the output dictionary defined in the interface of each instrument
    
    def __init__(self, app, mainwindow, parent, config_file, name_logger=__package__):
        # app           = The pyqt5 QApplication() object
        # mainwindow    = Main Window of the application
        # parent        = a QWidget (or QMainWindow) object that will be the parent for the gui of this device. Normally, this is an mdi object contained in the mainwindow object
        #                 This QWdiget must have an attribute called 'containers' which is a dictionary, containing the following keys
        #                 'logging','tabledata','instruments','plots'
        #                 Each item of parent.containers is a dictionary, and it identifies a different panel  
        #                 For each item, the sub-item with key 'container' is a widget that will host the relative content.
        #                   E.g. parent.containers['plots']['container'] is a QWidget that will host the plots
        # config_file   = a valid .json file containing all settings of this experiment
        # name_logger   = The name of the logger used for this particular experiment. If none is specified, the name of the package (i.e. ergastirio) is used as logger name

        self.mainwindow = mainwindow
        self.app = app
        self.parent = parent
        self.name_logger = name_logger #Setting this property will also create the logger,set defaulat output style, and store the logger object in self.logger (see @name_logger.setter)
        
        ergastirio.experiment_initializer.create_gui_logging(self,self.parent.containers['logging']['container']) #By calling this function here (instead than later inside set_up_experiment_gui() ) we make sure that any log message is shown in the GUI
                                                                                                                    #Need to implement this more elegantly
        self._refresh_time_internal_trigger = 0.1                #in s. Refresh time of internal trigger (when used)
        
        self.continous_acquisition = False      #Boolean variable, true when performing continuous acquisition (i.e. when internal trigger is used)
        self.triggered_acquisition = False      #Boolean variable, true when external trigger is used
        self._numb_acquisitions_per_trigger = 1 #Every time that a trigger is "fired" (either by internal trigger, or by external trigger, 
                                                #or by clicking on 'take single acqusitions') the software takes a number of acquisitions 
                                                #equal to self._numb_acquisitions_per_trigger, and separated in time by self._refresh_time_multiple_acquisitions
        self._refresh_time_multiple_acquisitions = 0.05 #in s 
        self.average_acquisition = True         #When this is set to true, after taking a number of acquisitions (give by self._numb_acquisitions_per_trigger )
                                                #it averages them and replace them by the average
        self.trigger_instrument = ''
        self.trigger_delay = 0

        self.config_file = config_file #Setting the config file name will also automatically open the file and load the settings (see @config_file.setter)
                                                                    
        self.mainwindow.experiment = self

        return

    @property
    def verbose(self):
        return self.verbose
    @verbose.setter
    def verbose(self,verbose):
        #When the verbose property of this interface is changed, we also update accordingly the level of the logger object
        if verbose: loglevel = logging.INFO
        else: loglevel = logging.CRITICAL
        self.logger.setLevel(level=loglevel)

    @property
    def name_logger(self):
        return self._name_logger
    @name_logger.setter
    def name_logger(self,name):
        #Create logger, and set default output style.
        self._name_logger = name
        self.logger = logging.getLogger(self._name_logger)
        self.verbose = self._verbose #This will automatically set the logger verbosity too
        if not self.logger.handlers:
            formatter = logging.Formatter(f"[{self.name_logger}]: %(message)s")
            ch = logging.StreamHandler()
            ch.setFormatter(formatter)
            self.logger.addHandler(ch)
        self.logger.propagate = False

    @property
    def config_file(self):
        return self._config_file
    @config_file.setter
    def config_file(self,config_file):
        self.logger.info(f"Setting the config file to {config_file}... [Note: this will reset the experiment]")
        self._config_file = config_file
        self.load_config()
        return 

    def load_config(self):
        '''
        Import experiment settings from the .json file stored in self.config_file, and then starts setting up the experiment 
        by calling the function ergastirio.experiment_initializer.setup(self).
        '''
        self.logger.info(f"Loading the content of {self.config_file}")
        try:
            with open(self.config_file) as jsonfile:
                self.config = json.load(jsonfile)
        except Exception as e:
            self.logger.error(f"An error occurred while loading the file. Fix the error and restart this application\n: {e}")
            return
        if not ergastirio.experiment_initializer.setup(self):
            return

    @property
    def refresh_time_internal_trigger(self):
        return self._refresh_time_internal_trigger
    @refresh_time_internal_trigger.setter
    def refresh_time_internal_trigger(self,r):
        try: 
            r = float(r)
            if self._refresh_time_internal_trigger == r: #in this case the number in the refresh time edit box is the same as the refresh time currently stored
                return True
        except ValueError:
            self.logger.error(f"The refresh time must be a valid number.")
            return self._refresh_time_internal_trigger
        if r < 0.001:
            self.logger.error(f"The refresh time must be positive and >= 1ms.")
            return self._refresh_time_internal_trigger
        self.logger.info(f"The refresh time (for internal trigger) is now {r} s.")
        self._refresh_time_internal_trigger = r
        return self._refresh_time_internal_trigger

    @property
    def refresh_time_multiple_acquisitions(self):
        return self._refresh_time_multiple_acquisitions
    @refresh_time_multiple_acquisitions.setter
    def refresh_time_multiple_acquisitions(self,r):
        try: 
            r = float(r)
            if self._refresh_time_multiple_acquisitions == r: #in this case the number in the refresh time edit box is the same as the refresh time currently stored
                return True
        except ValueError:
            self.logger.error(f"The refresh time must be a valid number.")
            return self._refresh_time_multiple_acquisitions
        if r < 0.001:
            self.logger.error(f"The refresh time must be positive and >= 1ms.")
            return self._refresh_time_multiple_acquisitions
        self.logger.info(f"The refresh time (for multiple acquisitions) is now {r} s.")
        self._refresh_time_multiple_acquisitions = r
        return self.refresh_time_multiple_acquisitions

    @property
    def numb_acquisitions_per_trigger(self):
        return self._numb_acquisitions_per_trigger
    @numb_acquisitions_per_trigger.setter
    def numb_acquisitions_per_trigger(self,n):
        try: 
            n = int(n)
            if self._numb_acquisitions_per_trigger == n:
                return True
        except ValueError:
            self.logger.error(f"The number of acquisitions per trigger must be a positive integer.")
            return self._numb_acquisitions_per_trigger
        if n < 1:
            self.logger.error(f"The number of acquisitions per trigger must be a positive integer.")
            return self._numb_acquisitions_per_trigger
        self.logger.info(f"The number of acquisitions per trigger is now {n}.")
        self._numb_acquisitions_per_trigger = n
        return self.numb_acquisitions_per_trigger

    def read_current_data_from_all_instruments(self):
        '''
        It looks into all the instruments interfaces (defined via the key 'interface' in each dictionary of the list exp.instruments) 
        and it extracts data from each instrument. From each instrument, the data to exctract is contained in the dictionary exp.instruments[i]['interface'].output
        '''
        current_data = []
        self.logger.info(f"Reading data from all instruments...")
        for instrument in self.instruments:        
            for data in instrument['interface'].output.values():
                current_data.append(data)  
        
        return current_data

    def store_current_data_from_all_instruments(self):
        '''
        '''
        current_data = self.read_current_data_from_all_instruments() #Read the current data from all instruments
        acq_numb = len(self.data) + 1 #We look at the number of rows of self.data
        now = datetime.datetime.now()
        time_string=now.strftime("%Y-%m-%d %H:%M:%S.%f")
        timestamp = datetime.datetime.timestamp(now)
        self.logger.info(f"[{time_string}] Acquisition #{acq_numb}")# {current_data}")
        current_data.insert(0, time_string)
        current_data.insert(0, timestamp )
        
        if ('PostProcessedData' in self.config.keys()):
            variables = {self.data_headers[i]: current_data[i] for i in range(len(current_data))}  #Create a dictionary with the values of the data that we just acquired
            for post_processed_data_name,post_processed_data_formula in self.config['PostProcessedData'].items():
                PostProcessedData = ergastirio.utils.evaluate_formula(post_processed_data_formula,variables)
                current_data.append(PostProcessedData)
        self.data.append(current_data)
        self.data_std.append([0]*(len(current_data)-2)) #Add a row of zeros to the standard deviations. The -2 is to avoid the columns corresponding to time and timestamp

    def set_trigger_modality(self,modality):
        if modality == 'internal':
            self.triggered_acquisition = False
            self.logger.info(f"Set the experiment to 'internal trigger modality'.")
        if modality == 'external':
            self.triggered_acquisition = True
            self.logger.info(f"Set the experiment to 'external trigger modality'.")

    def set_average_modality(self,modality):
        if modality == True:
            self.average_acquisition = True
            self.logger.info(f"Set averaged modality to True.")
        if modality == False:
            self.average_acquisition = False
            self.logger.info(f"Set averaged modality to False.")

    def take_single_acquisition(self):
        self.store_current_data_from_all_instruments()

    def take_single_set_of_acquisitions(self,counter):
        ''' Takes a sequence of acquisitions. The number of acquisitions is counter, separated by a time self.refresh_time_multiple_acquisitions.
            This function calls take_single_acquisition() and then calls itself after a time specified by self.refresh_time_multiple_acquisitions, 
            and with counter = counter -1.
            Normally, this function is called by another function with the input argument counter = self.numb_acquisitions_per_trigger
        '''
        self.take_single_acquisition()
        if counter > 1: #if there are more acquisitions to take we call the function again
            QtCore.QTimer.singleShot(int(self.refresh_time_multiple_acquisitions*1e3), lambda: self.take_single_set_of_acquisitions(counter=counter-1))
        elif self.average_acquisition and self.numb_acquisitions_per_trigger>1:
            self.logger.info(f"Averaging the last {self.numb_acquisitions_per_trigger} acquisitions and calculating errors...")
            self.make_average_last_acquisitions()

    def make_average_last_acquisitions(self):
        '''
        Take the last N elements of self.data (where N = self.numb_acquisitions_per_trigger) and replace them with their average (column-wise)
        It also populates the standard deviations (TO DO)
        '''
        N = self.numb_acquisitions_per_trigger
        rows_to_average = self.data[-N:] #Get the last N rows
        timestamp = rows_to_average[0][0]
        time = rows_to_average[0][1]
        rows_to_average2 = [elem[2:] for elem in rows_to_average] #Remove first two columns, which is timestamp and time
        averaged = (np.nanmean(np.array(rows_to_average2),axis=0)).tolist()
        std = (np.nanstd(np.array(rows_to_average2),axis=0)).tolist()
        averaged.insert(0, time)
        averaged.insert(0, timestamp )
        del self.data[-N:]      #Remove last N rows from data
        del self.data_std[-N:]  #Remove last N rows from data_std
        self.data.append(averaged)
        self.data_std.append(std)

    def trigger_single_set_acquisition(self):
        self.take_single_set_of_acquisitions(counter=self.numb_acquisitions_per_trigger)

    def update(self):
        '''
        If we are doing a continuous acquisition, it calls self.store_current_data_from_all_instruments(). If we are not in external triggered modality, it calls
        itself after a time equal to int(self.refresh_time_internal_trigger*1e3) (in ms)
        '''
        if self.continous_acquisition:
            self.take_single_set_of_acquisitions(counter=self.numb_acquisitions_per_trigger)
            if self.continous_acquisition == True and self.triggered_acquisition  == False:
                QtCore.QTimer.singleShot(int(self.refresh_time_internal_trigger*1e3), self.update)  
                
    def start_continous_acquisition(self):
        self.continous_acquisition = True
        if self.triggered_acquisition == True:
            self.logger.info(f"Waiting for next trigger from {self.trigger_instrument}...")
        else:
            self.update()
        return

    def stop_continous_acquisition(self):
        self.continous_acquisition = False

    def trigger(self):
        '''
        This function is called by instruments if they are set as trigger and when they fire a trigger
        '''
        if self.continous_acquisition:
            if self.triggered_acquisition:
                self.logger.info(f"Received trigger from {self.trigger_instrument}...")
                self.update()
                self.logger.info(f"Waiting for next trigger from {self.trigger_instrument}...")
            else:
                self.logger.info(f"Received a trigger from {self.trigger_instrument}, but experiment is not in 'external trigger' modality.")
        else:
            self.logger.info(f"Received a trigger from {self.trigger_instrument}, but experiment is not currently acquiring data.")

    def set_trigger_instrument(self,instrument_name,delay):
        try: 
            delay = float(delay)
        except ValueError:
            self.logger.error(f"The trigger must be a valid number.")
            return False
        if delay < 0:
            self.logger.error(f"The delay time must be positive and >= 1ms.")
            return False
        if (self.trigger_instrument == instrument_name) and (self.trigger_delay == delay): #in this case the number in the refresh time edit box is the same as the refresh time currently stored
            return True
        self.trigger_instrument = ''
        self.logger.info(f"Setting the instrument {instrument_name} as trigger, with delay {delay}...")
        for instrument in self.instruments:        
            if instrument['fullname'] == instrument_name:
                try:
                    instrument['interface'].set_trigger(self.trigger,delay=delay)
                    self.trigger_instrument = instrument_name
                    self.trigger_delay = delay
                except Exception as e:
                    self.logger.error(f"An error occurred while setting {self.trigger_instrument} as trigger:\n {e}")
                    return False
            else: #we set the trigger of this instrument to None
                try:
                    instrument['interface'].set_trigger(None,delay=delay)
                except Exception as e:
                    self.logger.error(f"{e}")
        if self.trigger_instrument == '':
            self.logger.error(f"There is not an instrument called {instrument_name}")
            return False
        self.logger.info(f"{self.trigger_instrument} was succesfully set as trigger.")
        return True

    def delete_current_data(self):
        self.logger.info(f"All store data was deleted.")
        self.data.clear()
        self.data_std.clear()

    def delete_row_from_data(self,row):
        try:
            self.data.pop(row)
            self.data_std.pop(row)
            if row == -1:
                row = 'Last'
            self.logger.info(f"{row} row has been removed from data.")
        except:
            pass

    def save_stored_data(self,filename):
        '''
        Saves the values of all currently stored data on file. 
        The data are saved in a tabular form and delimited by a comma.
        '''
        d =','
        header = ''
        for h in self.data_headers:
            header = header + h + d 
        for h in self.data_headers: #We skip the first element of headers which corresponds to acquisition time
            if h not in ['timestamp', 'time']: #This avoids generating the headers timestamp_std and time_std
                header = header + (h+'_std') +  d

        A = np.concatenate((np.array(self.data), np.array(self.data_std)),axis=1)
        print(A)
        np.savetxt(filename, A, delimiter=d, header=header, comments="",fmt='%s')#,fmt=d.join(['%s'] + ['%e']*(A.shape[1]-1)))
        self.logger.info(f"Saved all stored data (number of rows = {A.shape[0]} in the file {filename}")
        return

    def close_experiment(self):
        for instrument in self.instruments:
            try:
                instrument['interface'].close()
            except:
                pass

class EnhancedList(list):
    '''
    This class is used to generate an "event-based" list object. Everytime the content of the list changes, it also stores a copy of the list 
    in a certain property of all the objects specified in the list linked_objects. 
    Each element of linked_objects is a two-element list in the form [class_instance,class_property]. This behavior is useful when the linked objects
    are, e.g., plots and tables, and the targeted property is defined via a @setter, in order to automatically update parts of the gui

    Examples:

        class test():
            def __init__( self ):
                self.__value = "old value"

            @property
            def value( self ):
                return self.__value

            @value.setter
            def value( self, value ):
                self.__value = value
                print("Targeted list changed to " + str(value))

        a = EnhancedList([1,2,3])
        t=test()
        a.add_syncronized_objects([t,test.value])
        a.append(4)

    Targeted list changed to [1, 2, 3, 4]

    '''
    

    def __init__(self,  *args):
        self.linked_objects = []

    def add_syncronized_objects(self,list_objects):
        self.linked_objects.append(list_objects)

    def sync(self):
        for obj in self.linked_objects:
            obj[1].fset(obj[0], self.copy())

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        self.sync()
    def __delitem__(self, value):
        super().__delitem__(value)
        self.sync()
    def __add__(self, value):
        super().__add__(value)
        self.sync()
    def __iadd__(self, value):
        super().__iadd__(value)
        self.sync()
    def append(self, value):
        super().append(value)
        self.sync()
    def remove(self, value):
        super().remove(value)
        self.sync()
    def insert(self, *args):
        super().insert(*args)
        self.sync()
    def pop(self, *args):
        super().pop(*args)
        self.sync()
    def clear(self):
        super().clear()
        self.sync()